/*
 *  TotalOrder.scala
 *  (LucreData)
 *
 *  Copyright (c) 2011-2012 Hanns Holger Rutz. All rights reserved.
 *
 *  This software is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either
 *  version 2, june 1991 of the License, or (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License (gpl.txt) along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  For further information, please contact Hanns Holger Rutz at
 *  contact@sciss.de
 */

package de.sciss.collection
package txn

import de.sciss.lucre.{DataInput, DataOutput}
import txn.TotalOrder.Set.EmptyEntry
import de.sciss.lucre.stm._
import annotation.{switch, tailrec}


/**
 * A transactional data structure to maintain an ordered sequence of elements such
 * that two random elements can be compared in O(1).
 *
 * This uses an algorithm from the paper
 * Bender, M. and Cole, R. and Demaine, E. and Farach-Colton, M. and Zito, J.,
 * Two simplified algorithms for maintaining order in a list,
 * Algorithms—ESA 2002, pp. 219--223, 2002.
 *
 * The `relabel` method is based on the Python implementation by
 * David Eppstein, as published at http://www.ics.uci.edu/~eppstein/PADS/OrderedSequence.py
 * however a bug resulting in a relabel size of 1 was fixed.
 *
 * Original note: "Due to rebalancing on the integer tags used to maintain order,
 * the amortized time per insertion in an n-item list is O(log n)."
 */
object TotalOrder {
   private val SER_VERSION = 0

   // ---- Set ----

   object Set {
      def empty[ S <: Sys[ S ]]( implicit tx: S#Tx ) : Set[ S ] = empty()

      def empty[ S <: Sys[ S ]]( rootTag: Int = 0 )( implicit tx: S#Tx ) : Set[ S ] = {
         val id = tx.newID()
         new SetNew( id, rootTag, tx.newIntVar( id, 1 ), tx )
      }

      implicit def serializer[ S <: Sys[ S ]] : TxnSerializer[ S#Tx, S#Acc, Set[ S ]] =
         new SetSerializer[ S ] // ( relabelObserver )

      sealed trait EntryOption[ S <: Sys[ S ]] {
         protected type E     = Entry[ S ]
         protected type EOpt  = EntryOption[ S ] /* with MutableOption[ S ] */

//         def tag( implicit tx: S#Tx ) : Int
         private[Set] def tagOr( empty: Int )( implicit tx: S#Tx ) : Int

         private[Set] def updatePrev( e: EOpt )( implicit tx: S#Tx ) : Unit
         private[Set] def updateNext( e: EOpt )( implicit tx: S#Tx ) : Unit
         private[Set] def updateTag( value: Int )( implicit tx: S#Tx ) : Unit
         def orNull : E
         def isDefined: Boolean
         def isEmpty: Boolean
      }

      final class EmptyEntry[ S <: Sys[ S ]] private[TotalOrder]() extends EntryOption[ S ] /* with EmptyMutable */ {
         private[Set] def updatePrev( e: EOpt )( implicit tx: S#Tx ) {}
         private[Set] def updateNext( e: EOpt )( implicit tx: S#Tx ) {}
         def orNull : E = null
         private[Set] def updateTag( value: Int )( implicit tx: S#Tx ) {
            sys.error( "Internal error - shouldn't be here" )
         }
         private[Set] def tagOr( empty: Int )( implicit tx: S#Tx ) = empty
         def isDefined = false
         def isEmpty   = true
      }
//      case object EmptyEntry extends EmptyEntryLike

      final class Entry[ S <: Sys[ S ]] private[TotalOrder]( val id: S#ID, set: Set[ S ], tagVal: S#Var[ Int ],
                                                             prevRef: S#Var[ EntryOption[ S ] /* with MutableOption[ S ] */],
                                                             nextRef: S#Var[ EntryOption[ S ] /* with MutableOption[ S ] */])
      extends EntryOption[ S ] with Mutable[ S ] with Ordered[ S#Tx, Entry[ S ]] {

         def compare( that: Entry[ S ])( implicit tx: S#Tx ) : Int = {
            val thisTag = tag
            val thatTag = that.tag
            if( thisTag < thatTag ) -1 else if( thisTag > thatTag ) 1 else 0
         }

         def tag( implicit tx: S#Tx ) : Int = tagVal.get
         private[Set] def tagOr( empty: Int )( implicit tx: S#Tx ) = tagVal.get

         def prev( implicit tx: S#Tx ) : EOpt = prevRef.get
         def next( implicit tx: S#Tx ) : EOpt = nextRef.get
         private[Set] def prevOrNull( implicit tx: S#Tx ) : E = prevRef.get.orNull
         private[Set] def nextOrNull( implicit tx: S#Tx ) : E = nextRef.get.orNull
         def orNull : E = this
         def isDefined  = true
         def isEmpty    = false

         private[Set] def updatePrev( e: EOpt )( implicit tx: S#Tx ) { prevRef.set( e )}
         private[Set] def updateNext( e: EOpt )( implicit tx: S#Tx ) { nextRef.set( e )}
         private[Set] def updateTag( value: Int )( implicit tx: S#Tx ) { tagVal.set( value )}

         protected def writeData( out: DataOutput ) {
            tagVal.write( out )
            prevRef.write( out )
            nextRef.write( out )
         }

         protected def disposeData()( implicit tx: S#Tx ) {
            // then free the refs
            prevRef.dispose()
            nextRef.dispose()
            tagVal.dispose()
         }

         def remove()( implicit tx: S#Tx ) { set.remove( this )}
         def append()( implicit tx: S#Tx ) : E = set.insertAfter( this )
         def prepend()( implicit tx: S#Tx ) : E = set.insertBefore( this )

         def removeAndDispose()( implicit tx: S#Tx ) {
            remove()
            dispose()
         }

         def validate( msg: => String )( implicit tx: S#Tx ) {
            val recTag  = tag
            if( prev.isDefined ) {
               val prevTag = prev.orNull.tag
               assert( prevTag < recTag, "prev " + prevTag + " >= rec " + recTag + " - " + msg )
            }
            if( next.isDefined ) {
               val nextTag = next.orNull.tag
               assert( recTag < nextTag, "rec " + recTag+ " >= next " + nextTag + " - " + msg )
            }
         }
      }
   }

   private final class SetSerializer[ S <: Sys[ S ]] extends TxnSerializer[ S#Tx, S#Acc, Set[ S ]] {
      def read( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : Set[ S ] = {
         val id = tx.readID( in, access )
         val version = in.readUnsignedByte()
         require( version == SER_VERSION, "Incompatible serialized version (found " + version +
            ", required " + SER_VERSION + ")." )
         val sizeVal = tx.readIntVar( id, in )

         new SetRead[ S ]( id, sizeVal, in, access, tx )
      }

      def write( v: Set[ S ], out: DataOutput ) { v.write( out )}
   }

   private final class SetRead[ S <: Sys[ S ]]( val id: S#ID, protected val sizeVal: S#Var[ Int ],
                                                in: DataInput, access: S#Acc, tx0: S#Tx ) extends Set[ S ] {
//      val root = tx0.readVar[ Set.Entry[ S ]]( id, in )( EntrySerializer )
      val root = EntrySerializer.read( in, access )( tx0 )
   }

   private final class SetNew[ S <: Sys[ S ]]( val id: S#ID, rootTag: Int, protected val sizeVal: S#Var[ Int ], tx0: S#Tx )
   extends Set[ S ] {
      me =>

      val root: E = {
         val rootID  = tx0.newID()
         val tagVal  = tx0.newIntVar( rootID, rootTag )
         val prevRef = tx0.newVar[ EOpt ]( id, empty )( EntryOptionSerializer )
         val nextRef = tx0.newVar[ EOpt ]( id, empty )( EntryOptionSerializer )
         new E( rootID, me, tagVal, prevRef, nextRef )
      }
   }

   sealed trait Set[ S <: Sys[ S ]] extends TotalOrder[ S ] /* with Reader[ Set[ S ]#E ] */ {
      me =>

      type E     = Set.Entry[ S ]
      protected type EOpt  = Set.EntryOption[ S ] /* with MutableOption[ S ] */

      protected def sizeVal: S#Var[ Int ]
      protected val empty  = new Set.EmptyEntry[ S ]

      def root: E

      final def readEntry( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : E = EntrySerializer.read( in, access )

      protected implicit object EntrySerializer extends TxnSerializer[ S#Tx, S#Acc, E ] {
         def read( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : E = {
            val id      = tx.readID( in, access )
            val tagVal  = tx.readIntVar( id, in )
            val prevRef = tx.readVar[ EOpt ]( id, in )( EntryOptionSerializer )
            val nextRef = tx.readVar[ EOpt ]( id, in )( EntryOptionSerializer )
            new E( id, me, tagVal, prevRef, nextRef )
         }

         def write( v: E, out: DataOutput ) { v.write( out )}
      }

      protected implicit object EntryOptionSerializer extends TxnSerializer[ S#Tx, S#Acc, EOpt ] {
//         def read( in: DataInput ) : EOpt = system.readOptionMut[ EOpt ]( in )
         def read( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : EOpt = {
            (in.readUnsignedByte(): @switch) match {
               case 0 => me.empty
               case 1 => EntrySerializer.read( in, access )
               case cookie => sys.error( "Unexpected cookie " + cookie )
            }
         }

         def write( v: EOpt, out: DataOutput ) {
            val e = v.orNull
            if( e == null ) {
               out.writeUnsignedByte( 0 )
            } else {
               out.writeUnsignedByte( 1 )
               e.write( out )
            }
         }
      }

      final protected def disposeData()( implicit tx: S#Tx ) {
         root.dispose()
         sizeVal.dispose()
      }

      final protected def writeData( out: DataOutput ) {
         out.writeUnsignedByte( SER_VERSION )
         sizeVal.write( out )
         root.write( out )
      }

      final def insertAfter( prev: E )( implicit tx: S#Tx ) : E = {
         val next       = prev.next
         val nextTag    = next.tagOr( Int.MaxValue )
         val recID      = tx.newID()
         val recPrevRef = tx.newVar[ EOpt ]( recID, prev )
         val recNextRef = tx.newVar[ EOpt ]( recID, next )
         val prevTag    = prev.tag
         val recTag     = prevTag + ((nextTag - prevTag + 1) >>> 1)
         val recTagVal  = tx.newIntVar( recID, recTag )
         val rec        = new E( recID, this, recTagVal, recPrevRef, recNextRef )
         prev.updateNext( rec )
         next.updatePrev( rec )
         sizeVal.transform( _ + 1 )
         if( recTag == nextTag ) relabel( rec )
         rec
      }

      final def insertBefore( next: E )( implicit tx: S#Tx ) : E = {
         val prev       = next.prev
         val prevTag    = prev.tagOr( 0 )
         val recID      = tx.newID()
         val recPrevRef = tx.newVar[ EOpt ]( recID, prev )
         val recNextRef = tx.newVar[ EOpt ]( recID, next )
         val nextTag    = next.tag
         val recTag     = prevTag + ((nextTag - prevTag + 1) >>> 1)
         val recTagVal  = tx.newIntVar( recID, recTag )
         val rec        = new E( recID, this, recTagVal, recPrevRef, recNextRef )
         next.updatePrev( rec )
         prev.updateNext( rec )
         sizeVal.transform( _ + 1 )
         if( recTag == nextTag ) relabel( rec )
         rec
      }

      private[TotalOrder] final def remove( entry: E )( implicit tx: S#Tx ) {
         val p = entry.prev
         val n = entry.next
         p.updateNext( n )
         n.updatePrev( p )
         sizeVal.transform( _ - 1 )
      }

      final def size( implicit tx: S#Tx ) : Int = sizeVal.get

      final def head( implicit tx: S#Tx ) : E = {
         var e = root
         var p = e.prevOrNull
         while( p ne null ) {
            e = p
            p = p.prevOrNull
         }
         e
      }

      final def tagList( from: E )( implicit tx: S#Tx ) : List[ Int ] = {
         val b       = List.newBuilder[ Int ]
         var entry   = from
         while( entry ne null ) {
            b       += entry.tag
            entry    = entry.nextOrNull
         }
         b.result()
      }

      /**
       * Relabels from a this entry to clean up collisions with
       * its successors' tags.
       *
       * Original remark from Eppstein:
       * "At each iteration of the rebalancing algorithm, we look at
       * a contiguous subsequence of items, defined as the items for which
       * self._tag &~ mask == base.  We keep track of the first and last
       * items in the subsequence, and the number of items, until we find
       * a subsequence with sufficiently low density, at which point
       * we space the tags evenly throughout the available values.
       *
       * The multiplier controls the growth of the threshhold density;
       * it is 2/T for the T parameter described by Bender et al.
       * Large multipliers lead to fewer relabels, while small items allow
       * us to handle more items with machine integer tags, so we vary the
       * multiplier dynamically to allow it to be as large as possible
       * without producing integer overflows."
       */
      private def relabel( _first: E )( implicit tx: S#Tx ) {
         var mask       = -1
         var thresh     = 1.0
         var num        = 1
   //      val mul     = 2/((2*len(self))**(1/30.))
         val mul        = 2 / math.pow( size << 1, 1/30.0 )
         var first      = _first
         var last       = _first
         var base       = _first.tag
         do {
            var prev    = first.prevOrNull
            while( (prev ne null) && ((prev.tag & mask) == base) ) {
               first    = prev
               prev     = prev.prevOrNull
               num     += 1
            }
            var next    = last.nextOrNull
            while( (next ne null) && ((next.tag & mask) == base) ) {
               last     = next
               next     = next.nextOrNull
               num     += 1
            }
   //         val inc = (mask + 1) / num
            val inc = -mask / num

            // important: we found a corner case where _first is the last
            // element in the list with a value of 0x7FFFFFFF. in this
            // case, if the predecessor is smaller in value, the original
            // algorithm would immediately terminate with num == 1, which
            // will obviously leave the tag unchanged! thus we must add
            // the additional condition that num is greater than 1!
            if( (inc >= thresh) && (num > 1) ) {   // found rebalanceable range
//               observer.beforeRelabeling( first, num )
//sys.error( "TODO" )

   //            while( !(item eq last) ) {
               // Note: this was probably a bug in Eppstein's code
               // -- it ran for one iteration less which made
               // the test suite fail for very dense tags. it
               // seems now it is correct with the inclusion
               // of last in the tag updating.
               next = first
               var cnt = 0; while( cnt < num ) {
                  next.updateTag( base )
                  next        = next.nextOrNull
                  base       += inc
                  cnt        += 1
               }
//sys.error( "TODO" )
//               observer.afterRelabeling( first, num )
               return
            }
            mask   <<= 1      // next coarse step
            base    &= mask
            thresh  *= mul
         } while( mask != 0 )
         sys.error( "label overflow" )
      }
   }

   // ---- Map ----

   object Map {
      def empty[ S <: Sys[ S ], A ]( relabelObserver: Map.RelabelObserver[ S#Tx, A ], entryView: A => Map.Entry[ S, A ],
                                     rootTag: Int = 0 )( implicit tx: S#Tx, keySerializer: TxnSerializer[ S#Tx, S#Acc, A ]) : Map[ S, A ] = {
         val id = tx.newID()
         new MapNew[ S, A ]( id, tx.newIntVar( id, 1 ), relabelObserver, entryView, rootTag, tx )
      }

      implicit def serializer[ S <: Sys[ S ], A ]( relabelObserver: Map.RelabelObserver[ S#Tx, A ], entryView: A => Map.Entry[ S, A ])
         ( implicit keySerializer: TxnSerializer[ S#Tx, S#Acc, A ]) : TxnSerializer[ S#Tx, S#Acc, Map[ S, A ]] =
         new MapSerializer[ S, A ]( relabelObserver, entryView )

      /**
       * A `RelabelObserver` is notified before and after a relabeling is taking place due to
       * item insertions. The iterator passed to it contains all the items which are relabelled,
       * excluding the one that has caused the relabelling action.
       *
       * Note that there is a tricky case, when an object creates more than one total order entry,
       * and then calls `placeBefore` or `placeAfter` successively on these entries: For the first
       * entry, the iterator will not contain the inserted element, but when the second entry is
       * inserted, the iterator will contain the first entry, with the potential of causing trouble
       * as the entry may be contained in an incompletely initialized object.
       *
       * For example, in the case of storing pre-head, pre-tail and post elements in two orders, make
       * sure that the pre-tail insertion comes last. Because this will happen:
       *
       * (1) pre-head, post and pre-tail entries created (their tags are -1)
       * (2) pre-head placed, may cause relabelling, but then will be excluded from the iterator
       * (3) post placed, may cause relabelling, but then will be excluded from the iterator
       * (4) pre-tail placed, may cause relabelling, and while the pre-tail view will be excluded
       * from the iterator, the previously placed pre-head '''will''' be included in the iterator,
       * showing the item with pre-tail tag of `-1` in `beforeRelabeling`, however, fortunately,
       * with assigned tag in `afterRelabeling`.
       */
      trait RelabelObserver[ Tx, -A ] {
         /**
          * This method is invoked right before relabelling starts. That is, the items in
          * the `dirty` iterator are about to be relabelled, but at the point of calling
          * this method the tags still carry their previous values.
          */
         def beforeRelabeling( /* inserted: A, */ dirty: Iterator[ Tx, A ])( implicit tx: Tx ) : Unit
         /**
          * This method is invoked right after relabelling finishes. That is, the items in
          * the `clean` iterator have been relabelled and the tags carry their new values.
          */
         def afterRelabeling( /* inserted: A, */ clean: Iterator[ Tx, A ])( implicit tx: Tx ) : Unit
      }

      final class NoRelabelObserver[ Tx, A ]
      extends RelabelObserver[ Tx, A ] {
         def beforeRelabeling( /* inserted: A, */ dirty: Iterator[ Tx, A ])( implicit tx: Tx ) {}
         def afterRelabeling(  /* inserted: A, */ clean: Iterator[ Tx, A ])( implicit tx: Tx ) {}
      }

      final class Entry[ S <: Sys[ S ], A ] private[TotalOrder](
         map: Map[ S, A ], val id: S#ID, tagVal: S#Var[ Int ], prevRef: S#Var[ KeyOption[ S, A ]],
         nextRef: S#Var[ KeyOption[ S, A ]])
      extends Mutable[ S ] with Ordered[ S#Tx, Entry[ S, A ]] {
         private type E    = Entry[ S, A ]
         private type KOpt = KeyOption[ S, A ]

         def tag( implicit tx: S#Tx ) : Int = tagVal.get

def validate( msg: => String )( implicit tx: S#Tx ) {
   val recTag  = tag
   if( prev.isDefined ) {
      val prevTag = map.entryView( prev.get ).tag
      assert( prevTag < recTag, "prev " + prevTag + " >= rec " + recTag + " - " + msg )
   }
   if( next.isDefined ) {
      val nextTag = map.entryView( next.get ).tag
      assert( recTag < nextTag, "rec " + recTag+ " >= next " + nextTag + " - " + msg )
   }
}

         override def toString() = "Entry" + id.toString

         private[TotalOrder] def prev( implicit tx: S#Tx ) : KOpt = prevRef.get
         private[TotalOrder] def next( implicit tx: S#Tx ) : KOpt = nextRef.get
//         private[TotalOrder] def prevOrNull( implicit tx: S#Tx ) : A = prevRef.get.orNull
//         private[TotalOrder] def nextOrNull( implicit tx: S#Tx ) : A = nextRef.get.orNull
//         def orNull : E = this

         private[TotalOrder] def updatePrev( e: KOpt )( implicit tx: S#Tx ) { prevRef.set( e )}
         private[TotalOrder] def updateNext( e: KOpt )( implicit tx: S#Tx ) { nextRef.set( e )}
         private[TotalOrder] def updateTag( value: Int )( implicit tx: S#Tx ) { tagVal.set( value )}

         // ---- Ordered ----

         def compare( that: E )( implicit tx: S#Tx ) : Int = {
            val thisTag = tag
            val thatTag = that.tag
            if( thisTag < thatTag ) -1 else if( thisTag > thatTag ) 1 else 0
         }

         protected def writeData( out: DataOutput ) {
            tagVal.write( out )
            prevRef.write( out )
            nextRef.write( out )
         }

         protected def disposeData()( implicit tx: S#Tx ) {
            // then free the refs
            prevRef.dispose()
            nextRef.dispose()
            tagVal.dispose()
         }

//         def append( key: A )( implicit tx: S#Tx ) : E = map.insertBetween( )
//
//         def prepend( key: A )( implicit tx: S#Tx ) : E = map.insertBefore( this, key )

         def remove()( implicit tx: S#Tx ) { map.remove( this )}

         def removeAndDispose()( implicit tx: S#Tx ) {
            remove()
            dispose()
         }
      }
   }

   private[TotalOrder] sealed trait KeyOption[ S <: Sys[ S ], A ] extends Writer {
      private type KOpt = KeyOption[ S, A ]

//         def tag( implicit tx: S#Tx ) : Int
//         def updatePrev( e: KOpt )( implicit tx: S#Tx ) : Unit
//         def updateNext( e: KOpt )( implicit tx: S#Tx ) : Unit
//         def updateTag( value: Int )( implicit tx: S#Tx ) : Unit
      def orNull : Map.Entry[ S, A ]
      def isDefined: Boolean
      def isEmpty: Boolean
      def get : A
   }

   private[TotalOrder] final class EmptyKey[ S <: Sys[ S ], A ]
   extends KeyOption[ S, A ] /* with EmptyMutable */ {
      private type KOpt = KeyOption[ S, A ]

//      def updatePrev( e: KOpt )( implicit tx: S#Tx ) {}
//      def updateNext( e: KOpt )( implicit tx: S#Tx ) {}
//      def updateTag( value: Int )( implicit tx: S#Tx ) { throw new NoSuchElementException( "EmptyKey.updateTag" )}
      def isDefined: Boolean = false
      def isEmpty: Boolean = true
      def get : A = throw new NoSuchElementException( "EmptyKey.get" )
      def tag( implicit tx: S#Tx ) = Int.MaxValue
      def orNull : Map.Entry[ S, A ] = null

      def write( out: DataOutput ) { out.writeUnsignedByte( 0 )}

      override def toString = "<empty>"
   }

   private[TotalOrder] final class DefinedKey[ S <: Sys[ S ], A ]( map: Map[ S, A ], val get: A )
   extends KeyOption[ S, A ] {
      def isDefined: Boolean = true
      def isEmpty: Boolean = false
      def orNull : Map.Entry[ S, A ] = map.entryView( get )

      def write( out: DataOutput ) {
         out.writeUnsignedByte( 1 )
         map.keySerializer.write( get, out )
      }

      override def toString = get.toString
   }

//   private[TotalOrder] type KeyOption[ S <: Sys[ S ], A ] = KeyOption[ S, A ] with MutableOption[ S ]

   private final class MapSerializer[ S <: Sys[ S ], A ]( relabelObserver: Map.RelabelObserver[ S#Tx, A ],
                                                          entryView: A => Map.Entry[ S, A ])
                                                        ( implicit keySerializer: TxnSerializer[ S#Tx, S#Acc, A ])
   extends TxnSerializer[ S#Tx, S#Acc, Map[ S, A ]] {
      def read( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : Map[ S, A ] = {
         val id = tx.readID( in, access )
         val version = in.readUnsignedByte()
         require( version == SER_VERSION, "Incompatible serialized version (found " + version +
            ", required " + SER_VERSION + ")." )
         val sizeVal = tx.readIntVar( id, in )

         new MapRead[ S, A ]( id, sizeVal, relabelObserver, entryView, in, access, tx )
      }

      def write( v: Map[ S, A ], out: DataOutput ) { v.write( out )}
   }

   private final class MapRead[ S <: Sys[ S ], A ](
      val id: S#ID, protected val sizeVal: S#Var[ Int ], protected val observer: Map.RelabelObserver[ S#Tx, A ],
      val entryView: A => Map.Entry[ S, A ], in: DataInput, access: S#Acc, tx0: S#Tx )(
      implicit private[TotalOrder] val keySerializer: TxnSerializer[ S#Tx, S#Acc, A ])
   extends Map[ S, A ] {

      val root = EntrySerializer.read( in, access )( tx0 )
   }

   private final class MapNew[ S <: Sys[ S ], A ](
      val id: S#ID, protected val sizeVal: S#Var[ Int ], protected val observer: Map.RelabelObserver[ S#Tx, A ],
      val entryView: A => Map.Entry[ S, A ], rootTag: Int, tx0: S#Tx )(
      implicit private[TotalOrder] val keySerializer: TxnSerializer[ S#Tx, S#Acc, A ])
   extends Map[ S, A ] {
      val root: E = {
         val rootID  = tx0.newID()
         val tagVal  = tx0.newIntVar( rootID, rootTag )
         val prevRef = tx0.newVar[ KOpt ]( rootID, emptyKey )
         val nextRef = tx0.newVar[ KOpt ]( rootID, emptyKey )
         new Map.Entry[ S, A ]( this, rootID, tagVal, prevRef, nextRef )
      }
   }

   private final class MapEntrySerializer[ S <: Sys[ S ], A ]( map: Map[ S, A ])
   extends TxnSerializer[ S#Tx, S#Acc, Map.Entry[ S, A ]] {

      private type E    = Map.Entry[ S, A ]
      private type KOpt = KeyOption[ S, A ]

      def read( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : E = {
         import map.keyOptionSer
         val id      = tx.readID( in, access )
         val tagVal  = tx.readIntVar( id, in )
         val prevRef = tx.readVar[ KOpt ]( id, in )
         val nextRef = tx.readVar[ KOpt ]( id, in )
         new Map.Entry[ S, A ]( map, id, tagVal, prevRef, nextRef )
      }

      def write( v: E, out: DataOutput ) { v.write( out )}
   }

   private final class KeyOptionSerializer[ S <: Sys[ S ], A ]( map: Map[ S, A ])
   extends TxnSerializer[ S#Tx, S#Acc, KeyOption[ S, A ]] {
      private type KOpt = KeyOption[ S, A ]

      def write( v: KOpt, out: DataOutput ) { v.write( out )}

      def read( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : KOpt = {
         if( in.readUnsignedByte() == 0 ) map.emptyKey else {
            val key = map.keySerializer.read( in, access )
            new DefinedKey( map, key )
         }
      }
   }

   /*
    * A special iterator used for the relabel observer.
    */
   private final class RelabelIterator[ S <: Sys[ S ], A ]( recOff: Int, num: Int, recE: Map.Entry[ S, A ], firstK: A,
                                                            entryView: A => Map.Entry[ S, A ])
   extends Iterator[ S#Tx, A ] {

      private var currK       = firstK
      private var cnt         = 0
      def hasNext : Boolean   = cnt < num

      def next()( implicit tx: S#Tx ) : A = {
         if( cnt == num ) throw new NoSuchElementException( "next on empty iterator" )
         val res     = currK
         cnt        += 1
         // if we're reaching the recE, skip it
         val currE   = if( cnt == recOff ) recE else entryView( currK )
         currK       = currE.next.get
         res
      }

      def reset() {
         currK = firstK
         cnt   = 0
      }
   }

   sealed trait Map[ S <: Sys[ S ], A ] extends TotalOrder[ S ] {
      map =>

      override def toString() = "TotalOrder.Map" + id.toString

      final type E                  = Map.Entry[ S, A ]
      final protected type KOpt     = KeyOption[ S, A ]

      final private[TotalOrder] val emptyKey: KOpt = new EmptyKey[ S, A ]
      final implicit val EntrySerializer: TxnSerializer[ S#Tx, S#Acc, E ] = new MapEntrySerializer[ S, A ]( this )
//      final private[TotalOrder] implicit val EntryOptionSerializer : MutableOptionReader[ S, KOpt ] = new MapEntryOptionReader[ S, A ]( this )
      final private[TotalOrder] implicit val keyOptionSer : TxnSerializer[ S#Tx, S#Acc, KOpt ] = new KeyOptionSerializer[ S, A ]( this )

      protected def sizeVal: S#Var[ Int ]
      protected def observer: Map.RelabelObserver[ S#Tx, A ]

      private[TotalOrder] def keySerializer: TxnSerializer[ S#Tx, S#Acc, A ]
      def entryView: A => E

      def root: E
//      final def max: EOpt = Empty  // grmpfff

      final def readEntry( in: DataInput, access: S#Acc )( implicit tx: S#Tx ) : E = EntrySerializer.read( in, access )

      final protected def disposeData()( implicit tx: S#Tx ) {
         root.dispose()
         sizeVal.dispose()
      }

      final protected def writeData( out: DataOutput ) {
         out.writeUnsignedByte( SER_VERSION )
         sizeVal.write( out )
         root.write( out )
      }

//      private[TotalOrder] def insertAfter( prevE: E, key: A )( implicit tx: S#Tx ) : E = {
//         val nextO      = prevE.next
//         if( nextO.isDefined ) {
//            entryView( nextO.get )
//         } else null
//         insertBetween( prevE, nextE, key )
//      }

//      private[TotalOrder] def insertBefore( nextE: E, key: A )( implicit tx: S#Tx ) : E = {
//         val prevO      = nextE.prev
//         val prevE      = if( prevO.isDefined ) entryView( prevO.get ) else null
//         insertBetween( prevE, nextE, key )
//      }

      def insert()( implicit tx: S#Tx ) : E = {
         val id            = tx.newID()
         val recTagVal     = tx.newIntVar( id, -1 )
         val recPrevRef    = tx.newVar[ KOpt ]( id, emptyKey )
         val recNextRef    = tx.newVar[ KOpt ]( id, emptyKey )
         new Map.Entry( this, id, recTagVal, recPrevRef, recNextRef )
      }

      def placeAfter( prev: A, key: A )( implicit tx: S#Tx ) {
         val prevE = entryView( prev )
         val nextO = prevE.next
         placeBetween( prevE, new DefinedKey[ S, A ]( map, prev ), nextO.orNull, nextO, key )
      }

      def placeBefore( next: A, key: A )( implicit tx: S#Tx ) {
         val nextE = entryView( next )
         val prevO = nextE.prev
         placeBetween( prevO.orNull, prevO, nextE, new DefinedKey[ S, A ]( map, next ), key )
      }

      private[TotalOrder] def placeBetween( prevE: E, prevO: KOpt, nextE: E, nextO: KOpt, key: A )( implicit tx: S#Tx ) {
         val prevTag       = if( prevE ne null ) prevE.tag else 0 // could use Int.MinValue+1, but that collides with Octree max space
         val nextTag       = if( nextE ne null ) nextE.tag else Int.MaxValue

assert( prevTag < nextTag, "placeBetween - prev is " + prevTag + ", while next is " + nextTag )

         val recTag        = prevTag + ((nextTag - prevTag + 1) >>> 1)
         val recE          = entryView( key ) // new Map.Entry( this, system.newID(), recTagVal, recPrevRef, recNextRef )

         require( recE.tag == -1 && prevTag >=0 && nextTag >= 0, {
            val msg = new StringBuilder()
            if( recE.tag != -1 ) msg.append( "Placed key was already placed before. " )
            if( prevTag < 0 )    msg.append( "Predecessor of placed key has not yet been placed. " )
            if( nextTag < 0 )    msg.append( "Successor of placed key has not yet been placed. " )
            msg.toString()
         })

//         val recTagVal     = system.newInt( recTag )
//         val recPrevRef    = system.newVal[ KOpt ]( prevO )
//         val recNextRef    = system.newVal[ KOpt ]( nextO )
         recE.updateTag( recTag )
         recE.updatePrev( prevO )
         recE.updateNext( nextO )
         val defK          = new DefinedKey[ S, A ]( this, key )
         if( prevE ne null ) prevE.updateNext( defK )
         if( nextE ne null ) nextE.updatePrev( defK )
         sizeVal.transform( _ + 1 )
         if( recTag == nextTag ) relabel( key, recE )
      }

//      def validate( key: A )( implicit tx: S#Tx ) : Boolean = {
//         val rec     = entryView( key )
//         val recTag  = rec.tag
//         val nextTag = rec.next.tag
//         val dirty   = recTag == nextTag
//         if( dirty ) relabel( key, rec )
//         dirty
//      }

      private[TotalOrder] def remove( e: E )( implicit tx: S#Tx ) {
         val p = e.prev
         val n = e.next
         if( p.isDefined ) p.orNull.updateNext( n )
         if( n.isDefined ) n.orNull.updatePrev( p )
         sizeVal.transform( _ - 1 )
      }

      final def size( implicit tx: S#Tx ) : Int = sizeVal.get

      final def head( implicit tx: S#Tx ) : E = {
         @tailrec def step( e: E ) : E = {
            val prevO = e.prev
            if( prevO.isEmpty ) e else step( prevO.orNull )
         }
         step( root )
      }

      final def tagList( from: E )( implicit tx: S#Tx ) : List[ Int ] = {
         val b = List.newBuilder[ Int ]
         @tailrec def step( e: E ) : List[ Int ] = {
            b += e.tag
            val nextO = e.next
            if( nextO.isEmpty ) b.result() else {
               step( nextO.orNull )
            }
         }
         step( from )
      }

      /*
       * Relabels from a this entry to clean up collisions with
       * its successors' tags.
       *
       * Original remark from Eppstein:
       * "At each iteration of the rebalancing algorithm, we look at
       * a contiguous subsequence of items, defined as the items for which
       * self._tag &~ mask == base.  We keep track of the first and last
       * items in the subsequence, and the number of items, until we find
       * a subsequence with sufficiently low density, at which point
       * we space the tags evenly throughout the available values.
       *
       * The multiplier controls the growth of the threshhold density;
       * it is 2/T for the T parameter described by Bender et al.
       * Large multipliers lead to fewer relabels, while small items allow
       * us to handle more items with machine integer tags, so we vary the
       * multiplier dynamically to allow it to be as large as possible
       * without producing integer overflows."
       */
      private def relabel( recK: A, recE: E )( implicit tx: S#Tx ) {
         var mask       = -1
         var thresh     = 1.0
         var num        = 1
   //      val mul     = 2/((2*len(self))**(1/30.))
         val mul        = 2 / math.pow( size << 1, 1/30.0 )
         var firstE     = recE
         var firstK     = recK
         var lastE      = recE
         var base       = recE.tag
         var recOff     = 0

         do {
            @tailrec def stepLeft() {
               val prevO = firstE.prev
               if( prevO.isDefined ) {
                  val prevK   = prevO.get
                  val prevE   = entryView( prevK )
                  if( (prevE.tag & mask) == base ) {
                     firstE   = prevE
                     firstK   = prevK
                     num     += 1
                     recOff  += 1
                     stepLeft()
                  }
               }
            }
            stepLeft()

            @tailrec def stepRight() {
               val nextO = lastE.next
               if( nextO.isDefined ) {
                  val nextE = entryView( nextO.get )
                  if( (nextE.tag & mask) == base ) {
                     lastE = nextE
                     num  += 1
                     stepRight()
                  }
               }
            }
            stepRight()

            if( num > 1 ) {
               val inc = -mask / num

               // important: we found a corner case where _first is the last
               // element in the list with a value of 0x7FFFFFFF. in this
               // case, if the predecessor is smaller in value, the original
               // algorithm would immediately terminate with num == 1, which
               // will obviously leave the tag unchanged! thus we must add
               // the additional condition that num is greater than 1!
               if( inc >= thresh ) {   // found rebalanceable range
                  val numM1 = num - 1
                  val relabelIter = if( recOff == 0 ) {
                     new RelabelIterator( -1, numM1, recE, firstE.next.get, entryView )
                  } else {
                     new RelabelIterator( recOff, numM1, /* if( recOff == numM1 ) numM1 else num, */ recE, firstK, entryView )
                  }
                  observer.beforeRelabeling( /* recK, */ relabelIter )

                  // Note: this was probably a bug in Eppstein's code
                  // -- it ran for one iteration less which made
                  // the test suite fail for very dense tags. it
                  // seems now it is correct with the inclusion
                  // of last in the tag updating.
                  var curr = firstE
                  var cnt = 0; while( cnt < num ) {
                     curr.updateTag( base )
                     val nextK   = curr.next.get
                     base       += inc
                     cnt        += 1
                     curr        = if( cnt == recOff ) recE else entryView( nextK )
                  }
                  relabelIter.reset()
                  observer.afterRelabeling( /* recK, */ relabelIter )
                  return
               }
            }
            mask   <<= 1      // next coarse step
            base    &= mask
            thresh  *= mul
         } while( mask != 0 )
         sys.error( "label overflow" )
      }
   }
}
sealed trait TotalOrder[ S <: Sys[ S ]] extends Mutable[ S ] {
   type E

   /**
    * The initial element created from which you can start to append and prepend.
    */
   def root : E // Entry[ S ]

   /**
    * Returns the head element of the structure. Note that this
    * is O(n) worst case.
    */
   def head( implicit tx: S#Tx ) : E // Entry[ S ]

   /**
    * The number of elements in the order. This is `1` for a newly
    * created order (consisting only of the root element).
    * You will rarely need this information except for debugging
    * purpose. The operation is O(1).
    */
   def size( implicit tx: S#Tx ) : Int

   def tagList( from: E )( implicit tx: S#Tx ) : List[ Int  ]
}
